@@ -1,15 +1,13 @@
 import asyncio
-import datetime
-import json
 import logging
 import os
 import subprocess
 from copy import deepcopy
-from hashlib import sha1
 
 import botocore.compat
+import dateutil.parser
 from botocore import UNSIGNED
-from botocore.compat import compat_shell_split
+from botocore.compat import compat_shell_split, total_seconds
 from botocore.config import Config
 from botocore.credentials import (
     _DEFAULT_ADVISORY_REFRESH_TIMEOUT,
@@ -27,6 +25,7 @@ from botocore.credentials import (
     CredentialResolver,
     CredentialRetrievalError,
     Credentials,
+    DeferredRefreshableCredentials,
     EnvProvider,
     InstanceMetadataProvider,
     InvalidConfigError,
@@ -39,18 +38,17 @@ from botocore.credentials import (
     RefreshableCredentials,
     RefreshWithMFAUnsupportedError,
     SharedCredentialProvider,
+    SSOCredentialFetcher,
     SSOProvider,
     SSOTokenLoader,
     UnauthorizedSSOTokenError,
     UnknownCredentialError,
-    _get_client_creator,
     _local_now,
     _parse_if_needed,
     _serialize_if_needed,
     parse,
     resolve_imds_endpoint_mode,
 )
-from dateutil.tz import tzutc
 
 from aiobotocore._helpers import resolve_awaitable
 from aiobotocore.config import AioConfig
@@ -58,6 +56,7 @@ from aiobotocore.tokens import AioSSOTokenProvider
 from aiobotocore.utils import (
     AioContainerMetadataFetcher,
     AioInstanceMetadataFetcher,
+    create_nested_client,
 )
 
 logger = logging.getLogger(__name__)
@@ -249,9 +248,22 @@ create_aio_mfa_serial_refresher = create_mfa_serial_refresher
 
 
 class AioCredentials(Credentials):
+    # Overrides for property accessors
+    def get_account_id(self):
+        return self.account_id
+
+    def get_access_key(self):
+        return self.access_key
+
+    def get_secret_key(self):
+        return self.secret_key
+
+    def get_token(self):
+        return self.token
+
     async def get_frozen_credentials(self):
         return ReadOnlyCredentials(
-            self.access_key, self.secret_key, self.token
+            self.access_key, self.secret_key, self.token, self.account_id
         )
 
 
@@ -260,6 +272,22 @@ class AioRefreshableCredentials(RefreshableCredentials):
         super().__init__(*args, **kwargs)
         self._refresh_lock = asyncio.Lock()
 
+    async def get_account_id(self):
+        await self._refresh()
+        return self._account_id
+
+    async def get_access_key(self):
+        await self._refresh()
+        return self._access_key
+
+    async def get_secret_key(self):
+        await self._refresh()
+        return self._secret_key
+
+    async def get_token(self):
+        await self._refresh()
+        return self._token
+
     # Redeclaring the properties so it doesn't call refresh
     # Have to redeclare setter as we're overriding the getter
     @property
@@ -267,7 +295,7 @@ class AioRefreshableCredentials(RefreshableCredentials):
         # TODO: this needs to be resolved
         raise NotImplementedError(
             "missing call to self._refresh. "
-            "Use get_frozen_credentials instead"
+            "Use get_frozen_credentials or get_access_key"
         )
         return self._access_key
 
@@ -280,7 +308,7 @@ class AioRefreshableCredentials(RefreshableCredentials):
         # TODO: this needs to be resolved
         raise NotImplementedError(
             "missing call to self._refresh. "
-            "Use get_frozen_credentials instead"
+            "Use get_frozen_credentials or get_secret_key instead"
         )
         return self._secret_key
 
@@ -293,7 +321,7 @@ class AioRefreshableCredentials(RefreshableCredentials):
         # TODO: this needs to be resolved
         raise NotImplementedError(
             "missing call to self._refresh. "
-            "Use get_frozen_credentials instead"
+            "Use get_frozen_credentials or get_token instead"
         )
         return self._token
 
@@ -301,6 +329,19 @@ class AioRefreshableCredentials(RefreshableCredentials):
     def token(self, value):
         self._token = value
 
+    @property
+    def account_id(self):
+        # TODO: this needs to be resolved
+        raise NotImplementedError(
+            "missing call to self._refresh. "
+            "Use get_frozen_credentials or get_account_id instead"
+        )
+        return self._account_id
+
+    @account_id.setter
+    def account_id(self, value):
+        self._account_id = value
+
     async def _refresh(self):
         if not self.refresh_needed(self._advisory_refresh_timeout):
             return
@@ -349,7 +390,7 @@ class AioRefreshableCredentials(RefreshableCredentials):
             return
         self._set_from_data(metadata)
         self._frozen_credentials = ReadOnlyCredentials(
-            self._access_key, self._secret_key, self._token
+            self._access_key, self._secret_key, self._token, self._account_id
         )
         if self._is_expired():
             msg = (
@@ -364,23 +405,21 @@ class AioRefreshableCredentials(RefreshableCredentials):
         return self._frozen_credentials
 
 
-class AioDeferredRefreshableCredentials(AioRefreshableCredentials):
+class AioDeferredRefreshableCredentials(
+    DeferredRefreshableCredentials, AioRefreshableCredentials
+):
     def __init__(self, refresh_using, method, time_fetcher=_local_now):
         self._refresh_using = refresh_using
         self._access_key = None
         self._secret_key = None
         self._token = None
+        self._account_id = None
         self._expiry_time = None
         self._time_fetcher = time_fetcher
         self._refresh_lock = asyncio.Lock()
         self.method = method
         self._frozen_credentials = None
 
-    def refresh_needed(self, refresh_in=None):
-        if self._frozen_credentials is None:
-            return True
-        return super().refresh_needed(refresh_in)
-
 
 class AioCachedCredentialFetcher(CachedCredentialFetcher):
     async def _get_credentials(self):
@@ -404,13 +443,16 @@ class AioCachedCredentialFetcher(CachedCredentialFetcher):
 
         creds = response['Credentials']
         expiration = _serialize_if_needed(creds['Expiration'], iso=True)
-        return {
+        credentials = {
             'access_key': creds['AccessKeyId'],
             'secret_key': creds['SecretAccessKey'],
             'token': creds['SessionToken'],
             'expiry_time': expiration,
+            'account_id': creds.get('AccountId'),
         }
 
+        return credentials
+
 
 class AioBaseAssumeRoleCredentialFetcher(
     BaseAssumeRoleCredentialFetcher, AioCachedCredentialFetcher
@@ -426,7 +468,9 @@ class AioAssumeRoleCredentialFetcher(
         kwargs = self._assume_role_kwargs()
         client = await self._create_client()
         async with client as sts:
-            return await sts.assume_role(**kwargs)
+            response = await sts.assume_role(**kwargs)
+            self._add_account_id_to_response(response)
+            return response
 
     async def _create_client(self):
         """Create an STS client using the source credentials."""
@@ -470,7 +514,9 @@ class AioAssumeRoleWithWebIdentityCredentialFetcher(
         # the token, explicitly configure the client to not sign requests.
         config = AioConfig(signature_version=UNSIGNED)
         async with self._client_creator('sts', config=config) as client:
-            return await client.assume_role_with_web_identity(**kwargs)
+            response = await client.assume_role_with_web_identity(**kwargs)
+            self._add_account_id_to_response(response)
+            return response
 
     def _assume_role_kwargs(self):
         """Get the arguments for assume role based on current configuration."""
@@ -503,6 +549,7 @@ class AioProcessProvider(ProcessProvider):
             secret_key=creds_dict['secret_key'],
             token=creds_dict.get('token'),
             method=self.METHOD,
+            account_id=creds_dict.get('account_id'),
         )
 
     async def _retrieve_credentials_using(self, credential_process):
@@ -533,6 +580,7 @@ class AioProcessProvider(ProcessProvider):
                 'secret_key': parsed['SecretAccessKey'],
                 'token': parsed.get('SessionToken'),
                 'expiry_time': parsed.get('Expiration'),
+                'account_id': self._get_account_id(parsed),
             }
         except KeyError as e:
             raise CredentialRetrievalError(
@@ -578,6 +626,7 @@ class AioEnvProvider(EnvProvider):
                     expiry_time,
                     refresh_using=fetcher,
                     method=self.METHOD,
+                    account_id=credentials['account_id'],
                 )
 
             return AioCredentials(
@@ -585,6 +634,7 @@ class AioEnvProvider(EnvProvider):
                 credentials['secret_key'],
                 credentials['token'],
                 method=self.METHOD,
+                account_id=credentials['account_id'],
             )
         else:
             return None
@@ -626,8 +676,13 @@ class AioSharedCredentialProvider(SharedCredentialProvider):
                     config, self.ACCESS_KEY, self.SECRET_KEY
                 )
                 token = self._get_session_token(config)
+                account_id = self._get_account_id(config)
                 return AioCredentials(
-                    access_key, secret_key, token, method=self.METHOD
+                    access_key,
+                    secret_key,
+                    token,
+                    method=self.METHOD,
+                    account_id=account_id,
                 )
 
 
@@ -648,8 +703,13 @@ class AioConfigProvider(ConfigProvider):
                     profile_config, self.ACCESS_KEY, self.SECRET_KEY
                 )
                 token = self._get_session_token(profile_config)
+                account_id = self._get_account_id(profile_config)
                 return AioCredentials(
-                    access_key, secret_key, token, method=self.METHOD
+                    access_key,
+                    secret_key,
+                    token,
+                    method=self.METHOD,
+                    account_id=account_id,
                 )
         else:
             return None
@@ -747,11 +807,14 @@ class AioAssumeRoleProvider(AssumeRoleProvider):
     async def _resolve_credentials_from_profile(self, profile_name):
         profiles = self._loaded_config.get('profiles', {})
         profile = profiles[profile_name]
-
         if (
             self._has_static_credentials(profile)
             and not self._profile_provider_builder
         ):
+            # This is only here for backwards compatibility. If this provider
+            # isn't given a profile provider builder we still want to be able
+            # to handle the basic static credential case as we would before the
+            # profile provider builder parameter was added.
             return self._resolve_static_credentials_from_profile(profile)
         elif self._has_static_credentials(
             profile
@@ -770,7 +833,6 @@ class AioAssumeRoleProvider(AssumeRoleProvider):
                     error_msg=error_message % profile_name,
                 )
             return credentials
-
         return await self._load_creds_via_assume_role(profile_name)
 
     def _resolve_static_credentials_from_profile(self, profile):
@@ -923,6 +985,7 @@ class AioContainerProvider(ContainerProvider):
             method=self.METHOD,
             expiry_time=_parse_if_needed(creds['expiry_time']),
             refresh_using=fetcher,
+            account_id=creds.get('account_id'),
         )
 
     def _create_fetcher(self, full_uri, *args, **kwargs):
@@ -944,6 +1007,7 @@ class AioContainerProvider(ContainerProvider):
                 'secret_key': response['SecretAccessKey'],
                 'token': response['Token'],
                 'expiry_time': response['Expiration'],
+                'account_id': response.get('AccountId'),
             }
 
         return fetch_creds
@@ -971,52 +1035,9 @@ class AioCredentialResolver(CredentialResolver):
         return None
 
 
-class AioSSOCredentialFetcher(AioCachedCredentialFetcher):
-    _UTC_DATE_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
-
-    def __init__(
-        self,
-        start_url,
-        sso_region,
-        role_name,
-        account_id,
-        client_creator,
-        token_loader=None,
-        cache=None,
-        expiry_window_seconds=None,
-        token_provider=None,
-        sso_session_name=None,
-    ):
-        self._client_creator = client_creator
-        self._sso_region = sso_region
-        self._role_name = role_name
-        self._account_id = account_id
-        self._start_url = start_url
-        self._token_loader = token_loader
-        self._token_provider = token_provider
-        self._sso_session_name = sso_session_name
-        super().__init__(cache, expiry_window_seconds)
-
-    def _create_cache_key(self):
-        args = {
-            'roleName': self._role_name,
-            'accountId': self._account_id,
-        }
-        if self._sso_session_name:
-            args['sessionName'] = self._sso_session_name
-        else:
-            args['startUrl'] = self._start_url
-
-        args = json.dumps(args, sort_keys=True, separators=(',', ':'))
-        argument_hash = sha1(args.encode('utf-8')).hexdigest()
-        return self._make_file_safe(argument_hash)
-
-    def _parse_timestamp(self, timestamp_ms):
-        # fromtimestamp expects seconds so: milliseconds / 1000 = seconds
-        timestamp_seconds = timestamp_ms / 1000.0
-        timestamp = datetime.datetime.fromtimestamp(timestamp_seconds, tzutc())
-        return timestamp.strftime(self._UTC_DATE_FORMAT)
-
+class AioSSOCredentialFetcher(
+    SSOCredentialFetcher, AioCachedCredentialFetcher
+):
     async def _get_credentials(self):
         """Get credentials by calling SSO get role credentials."""
         config = Config(
@@ -1028,7 +1049,16 @@ class AioSSOCredentialFetcher(AioCachedCredentialFetcher):
                 initial_token_data = self._token_provider.load_token()
                 token = (await initial_token_data.get_frozen_token()).token
             else:
-                token = self._token_loader(self._start_url)['accessToken']
+                token_dict = self._token_loader(self._start_url)
+                token = token_dict['accessToken']
+
+                # raise an UnauthorizedSSOTokenError if the loaded legacy token
+                # is expired to save a call to GetRoleCredentials with an
+                # expired token.
+                expiration = dateutil.parser.parse(token_dict['expiresAt'])
+                remaining = total_seconds(expiration - self._time_fetcher())
+                if remaining <= 0:
+                    raise UnauthorizedSSOTokenError()
 
             kwargs = {
                 'roleName': self._role_name,
@@ -1050,6 +1080,7 @@ class AioSSOCredentialFetcher(AioCachedCredentialFetcher):
                     'Expiration': self._parse_timestamp(
                         credentials['expiration']
                     ),
+                    'AccountId': self._account_id,
                 },
             }
             return credentials
@@ -1080,3 +1111,20 @@ class AioSSOProvider(SSOProvider):
             method=self.METHOD,
             refresh_using=sso_fetcher.fetch_credentials,
         )
+
+
+def _get_client_creator(session, region_name):
+    """Create a client creator function for use in credential providers.
+
+    This is the async version of botocore.credentials._get_client_creator that
+    uses aiobotocore's create_nested_client.
+    """
+
+    def client_creator(service_name, **kwargs):
+        create_client_kwargs = {'region_name': region_name}
+        create_client_kwargs.update(**kwargs)
+        return create_nested_client(
+            session, service_name, **create_client_kwargs
+        )
+
+    return client_creator
