@@ -24,6 +24,7 @@ from botocore.utils import (
     IMDSRegionProvider,
     InstanceMetadataFetcher,
     InstanceMetadataRegionFetcher,
+    PluginContext,
     ReadTimeoutError,
     S3ExpressIdentityCache,
     S3ExpressIdentityResolver,
@@ -31,7 +32,9 @@ from botocore.utils import (
     S3RegionRedirectorv2,
     get_environ_proxies,
     os,
+    reset_plugin_context,
     resolve_imds_endpoint_mode,
+    set_plugin_context,
 )
 
 import aiobotocore.httpsession
@@ -461,6 +464,10 @@ class AioS3RegionRedirectorv2(S3RegionRedirectorv2):
             0
         ].status_code in (301, 302, 307)
         is_permanent_redirect = error_code == 'PermanentRedirect'
+        is_opt_in_region_redirect = (
+            error_code == 'IllegalLocationConstraintException'
+            and operation.name != 'CreateBucket'
+        )
         if not any(
             [
                 is_special_head_object,
@@ -468,6 +475,7 @@ class AioS3RegionRedirectorv2(S3RegionRedirectorv2):
                 is_permanent_redirect,
                 is_special_head_bucket,
                 is_redirect_status,
+                is_opt_in_region_redirect,
             ]
         ):
             return
@@ -478,16 +486,21 @@ class AioS3RegionRedirectorv2(S3RegionRedirectorv2):
 
         if new_region is None:
             logger.debug(
-                f"S3 client configured for region {client_region} but the "
-                f"bucket {bucket} is not in that region and the proper region "
-                "could not be automatically determined."
+                "S3 client configured for region %s but the "
+                "bucket %s is not in that region and the proper region "
+                "could not be automatically determined.",
+                client_region,
+                bucket,
             )
             return
 
         logger.debug(
-            f"S3 client configured for region {client_region} but the bucket {bucket} "
-            f"is in region {new_region}; Please configure the proper region to "
-            f"avoid multiple unnecessary redirects and signing attempts."
+            "S3 client configured for region %s but the bucket %s "
+            "is in region %s; Please configure the proper region to "
+            "avoid multiple unnecessary redirects and signing attempts.",
+            client_region,
+            bucket,
+            new_region,
         )
         # Adding the new region to _cache will make construct_endpoint() to
         # use the new region as value for the AWS::Region builtin parameter.
@@ -614,16 +627,21 @@ class AioS3RegionRedirector(S3RegionRedirector):
 
         if new_region is None:
             logger.debug(
-                f"S3 client configured for region {client_region} but the bucket {bucket} is not "
+                "S3 client configured for region %s but the bucket %s is not "
                 "in that region and the proper region could not be "
-                "automatically determined."
+                "automatically determined.",
+                client_region,
+                bucket,
             )
             return
 
         logger.debug(
-            f"S3 client configured for region {client_region} but the bucket {bucket} is in region"
-            f" {new_region}; Please configure the proper region to avoid multiple "
-            "unnecessary redirects and signing attempts."
+            "S3 client configured for region %s but the bucket %s is in region"
+            " %s; Please configure the proper region to avoid multiple "
+            "unnecessary redirects and signing attempts.",
+            client_region,
+            bucket,
+            new_region,
         )
         endpoint = self._endpoint_resolver.resolve('s3', new_region)
         endpoint = endpoint['endpoint_url']
@@ -742,3 +760,35 @@ class AioContainerMetadataFetcher(ContainerMetadataFetcher):
                 f"container metadata: {e}"
             )
             raise MetadataRetrievalError(error_msg=error_msg)
+
+
+@contextlib.asynccontextmanager
+async def create_nested_client(session, service_name, **kwargs):
+    """Create a nested client with plugin context disabled.
+
+    If a client is created from within a plugin based on the environment variable,
+    an infinite loop could arise. Any clients created from within another client
+    must use this method to prevent infinite loops.
+
+    This is the async version of botocore.utils.create_nested_client that works
+    with aiobotocore's async session.
+
+    Usage:
+        async with create_nested_client(session, 'sts', region_name='us-east-1') as client:
+            response = await client.assume_role(...)
+    """
+    # Set plugin context to disabled
+    ctx = PluginContext(plugins="DISABLED")
+    token = set_plugin_context(ctx)
+
+    try:
+        # Create client context
+        async with session.create_client(service_name, **kwargs) as client:
+            # Reset plugin context immediately after client creation, matching botocore behavior
+            reset_plugin_context(token)
+            token = None
+
+            yield client
+    finally:
+        if token:
+            reset_plugin_context(token)
